/***************************************************************************
                          hashvector.h  -  class implementing hash/vector
                             -------------------
    begin                : Sun Dec 22 2002
    copyright            : (C) 2002 by Meilof
    email                : meilof@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __HASHVECTOR_H
#define __HASHVECTOR_H

#include <poslib/poslib.h>
#include "vector.h"
#include "pbl.h"

template<class TYPE> class vitem {
 public:
  vitem(const char *_name, TYPE _data) {
    name = strdup(_name);
    data = _data;
  }
  ~vitem() {
    free(name);
    if (data) delete data;
  }
  char *name;
  TYPE data;
};

typedef bool(*iteratorfn)(void *userdata, char *name, void *item);

#define HV_TRESHOLD 32

template<class TYPE> class hashvector {
 public:
  union {
    pvector<vitem<TYPE> *> *vec;
    pblHashTable_t *table;
  } data;
  int nitems;
  
  hashvector() {
    nitems = 0;
    data.vec = NULL;
  }

  void clean(bool dodelete = true) {
    if (nitems < HV_TRESHOLD) {
      if (data.vec) {
        data.vec->destroy(dodelete);
        delete data.vec;
        data.vec = NULL;
      }
    } else {
      TYPE dat;
      for(dat = (TYPE)pblHtFirst(data.table); dat; dat = (TYPE)pblHtNext(data.table)) {
        if (dodelete) delete dat;
        pblHtRemove(data.table, 0, 0);
      }
      pblHtDelete(data.table);
      data.table = NULL;
    }
    nitems = 0;
  }    

  ~hashvector() {
    clean(true);
  }

  void add_item(const char *name, TYPE dat) {
    if (nitems < HV_TRESHOLD - 1) {
      if (nitems == 0) data.vec = new pvector<vitem<TYPE> *>();
      data.vec->add(new vitem<TYPE>(name, dat));
      nitems++;
    } else if (nitems == HV_TRESHOLD - 1) {
      pblHashTable_t *table = pblHtCreate();
      vitem<TYPE> *v;
      for (int x = 0; x < data.vec->count; x++) {
        v = (*data.vec)[x];
        pblHtInsert(table, (void *)v->name, strlen(v->name), v->data);
        v->data = NULL;
        delete v;
        (*data.vec)[x] = NULL;
      }
      pblHtInsert(table, (void *)name, strlen(name), dat);
      delete data.vec;
      data.table = table;
      nitems++;
    } else {
      pblHtInsert(data.table, (void *)name, strlen(name), dat);
      nitems++;
    }
  }

  void convertback() {
    pvector<vitem<TYPE> *> *vec;
    vec = new pvector<vitem<TYPE> *>();
    TYPE dat;

    for(dat = (TYPE)pblHtFirst(data.table); dat; dat = (TYPE)pblHtNext(data.table)) {
      vec->add(new vitem<TYPE>((char *)pblHtCurrentKey(data.table), dat));
      pblHtRemove(data.table, 0, 0);
    }
    pblHtDelete(data.table);
    data.vec = vec;
  }     
  
  void remove_item(const char *name) {
    if (nitems == 0) return;
    if (nitems <= HV_TRESHOLD + 1) {
      if (nitems == HV_TRESHOLD + 1) convertback();
      for (int x = data.vec->count - 1; x >= 0; x--)
        if (strcmpi((*data.vec)[x]->name, name) == 0) { nitems--; data.vec->removeindex(x); }
    } else {
      TYPE dat = (TYPE)pblHtLookup(data.table, (void *)name, strlen(name));
      if (dat) {
        delete dat;
        pblHtRemove(data.table, 0, 0 );
        nitems--;
      }
    }
  }
  
  TYPE get_item(const char *name) {
    if (nitems == 0) return NULL;
    if (nitems <= HV_TRESHOLD) {
      for (int x = data.vec->count - 1; x >= 0; x--)
        if (strcmpi((*data.vec)[x]->name, name) == 0) return (*data.vec)[x]->data;
      return NULL;
    } else return (TYPE)pblHtLookup(data.table, (void *)name, strlen(name));
  }
  
  void iterate(iteratorfn fn, void *arg = NULL) {
    if (nitems == 0) return;
    if (nitems <= HV_TRESHOLD) {
      for (int x = 0; x < data.vec->count; x++)
        fn(arg, (*data.vec)[x]->name, (*data.vec)[x]->data);
    } else {
      TYPE dat;
      for(dat = (TYPE)pblHtFirst(data.table); dat; dat = (TYPE)pblHtNext(data.table)) {
        fn(arg, (char *)pblHtCurrentKey(data.table), dat);
      }
    }
  }
  
  bool isempty() {
    return (nitems == 0);
  }
};



//
//template<class TYPE> class tptr {
// public:
//  tptr(tptr<TYPE>& t) { data = t.data; }
//  tptr(TYPE *t) { data = t; }
//  tptr<TYPE> &operator=(tptr<TYPE>& t) { data = t.data; return *this; }
//  ~tptr() { delete data; }
//  TYPE *data;
//};
//
//
//#define hash_map_t hash_map<const char *, tptr<TYPE>, hash<const char *>, eqstr>
//#define pvector_t pvector<vitem<TYPE> *>
//
//struct eqstr {
//  bool operator()(const char* s1, const char* s2) const {
//    return strcmp(s1, s2) == 0;
//  }
//};
//
//template<class TYPE> class hashvector {
// public:
//  hashvector() {
//    nitems = 0;
//    hsv = NULL;
//    is_hash = false;
//  }
//  ~hashvector() {
//    if (is_hash) {
//      delete (hash_map_t *)hsv;
//    } else {
//      delete (pvector_t *)hsv;
//    }
//  }
//  void add_item(char *name, TYPE *t) {
//    nitems++;
//    if (!hsv)
//      hsv = new pvector_t();
//    if (!is_hash) {
//      if (nitems < TRESHOLD) {
//        ((pvector_t *)hsv)->add(new vitem<TYPE>(name, t));
//      } else {
//         hash_map_t *t = new hash_map_t();
//         vitem<TYPE> *i;
//         int x;
//         for (x = 0; x < ((pvector_t *)hsv)->count; x++) {
//           i = (*(pvector_t *)hsv)[x];
//           t->insert(value_type(i->name, tptr<TYPE>(i->data)));
//           i->data = NULL;
//         }
//         delete (pvector_t *)hsv;
//         hsv = t;
//         is_hash = true;
//      }
//    } else {
//      /* just add */
//      (*(hash_map_t *)hsv)[name] = new tptr<TYPE>(t);
//    }
//  }
//  void remove_item(char *name) {
//    if (is_hash) {
//      ((hash_map_t *)hsv)->erase(name);
//    } else {
//      for (int x = 0; x < nitems; x++) {
//        if (strcmpi(name, (*(pvector_t *)ptr)[x]->name) == 0)
//          ((pvector_t *)hsv)->removeindex(x);
//      }
//    }
//  }
//  TYPE *get_item(char *name) {
//    if (is_hash) {
//      hash_map_t::iterator it = ((hash_map_t *)hsv)->find(name);
//      if (it == ((hash_map_t *)hsv)->end()) return NULL;
//      else return it->data;
//    } else {
//      for (int x = 0; x < nitems; x++) {
//        if (strcmpi(name, (*(pvector_t *)hsv)[x]->name) == 0)
//          return (*(pvector_t *)hsv)[x]->data;
//      }
//    }
//  }
// private:
//  bool is_hash;
//  int nitems;
//  void *hsv;
//};

#endif /* __HASHVECTOR_H */
